
;////////////////////////////////////////////////////////////
;
; CAPITOLUL 5
; 
; 
; Un exemplu simplu de motor de inferenta backward chaining
; Fisier eng.lsp
;
;////////////////////////////////////////////////////////////



;////// Reprezentarea si manipularea regulilor ///////
;
; regula = (identif if lista-premize then concluzie)

(defun make-regula (id premize concluzie)
   "Construieste o regula"
   (list id 'if premize 'then concluzie))

(defun regula-premize (regula)
   "Intoarce premizele unei reguli"
   (nth 2 regula))

(defun regula-concluzie (regula)
   "Intoarce concluzia unei reguli"
   (nth 4 regula))

(defun regula-id (regula)
   "Intoarce identificatorul unei reguli"
   (nth 0 regula))

(defun regula-eq (regula1 regula2)
   "Functie de egalitate pentru reguli"
   (= (regula-id regula1)
      (regula-id regula2)))




;///////////  Functiile de rezolvare //////////////////////

(defun rezolva (concluzii)
   "Incearca sa demonstreze toate concluzii, folosind backward chaining"
   (format t "~% Rezolv ~a~%" concluzii)
   (cond ((null concluzii) t)
         ((member (car concluzii) *fapte*) (rezolva (cdr concluzii)))
         (t (and (rezolva-scop-simplu (car concluzii))
                 (rezolva (cdr concluzii))))))


(defun rezolva-scop-simplu (concluzie)
   "Incearca demonstrarea unei singure concluzii"
   (let ((reguli-aplicabile (find-reguli-aplicabile *reguli* concluzie)))
      (if (null reguli-aplicabile)
         nil
         (rezolva-scop-simplu-aux concluzie reguli-aplicabile))))

      
(defun rezolva-scop-simplu-aux (concluzie reguli-aplicabile)
   "Incearca demonstrarea concluzie-i pe baza regulilor si a faptelor"
   (if (null reguli-aplicabile)
      nil
      (let* ((regula-aleasa (alege-regula reguli-aplicabile))
             (rest-reguli (remove regula-aleasa reguli-aplicabile :test #'regula-eq)))
         (or (rezolva (regula-premize regula-aleasa))
             (rezolva-scop-simplu-aux concluzie rest-reguli)))))
   

   
(defun find-reguli-aplicabile (reguli concluzie)
   "Selecteaza toate regulile care au drept concluzie concluzie"
   (let ((apl-reguli ()))
      (dolist (r reguli apl-reguli)
         (if (member concluzie (regula-concluzie r))
            (setf apl-reguli (append (list r) apl-reguli))))))

(defun alege-regula (reguli)
   "Selecteaza o regula din lista celor ce pot fi aplicate
    Strategia de selectie = selecteaza prima regula gasita"
   (car reguli))



;///////////  Exemple  //////////////////////////////////


; Exemplu de set de reguli

(setf *reguli*
      (list
        (make-regula 1 '(a b c) '(d))
        (make-regula 2 '(e f g) '(h))
        (make-regula 3 '(i j) '(k))
        (make-regula 4 '(d h) '(l))
        ))


;Exemplu lista de fapte

(setf *fapte*
      '(a b c e f g))


;Exemplu de concluzie

(setf *concluzie* '(k))


;Exemplu de apel

(rezolva *concluzie*)
                           
              