
;; //////////////////////////////////////////////////////////////
;;
;;CAPITOL 8
;;
;; Generalizarea arborelui de explicatie
;; Fisier op-con.lsp
;;
;;//////////////////////////////////////////////////////////////

      
(defun constr-lista-subst-const (subst-l)
   (let ((rezultat nil))
      (dolist (pereche subst-l rezultat)
         (if (null (gaseste-legare (cdr pereche) rezultat))
            (setq rezultat (adauga-subst (cdr pereche)
                           (list 'var (gensym)) rezultat))))))

(defun constr-lista-subst-var (subst-l const-subst-list)
   (let ((rezultat nil))
      (dolist (pereche subst-l rezultat)
         (if (null (gaseste-legare (car pereche) rezultat))
            (setq rezultat (adauga-subst (car pereche)
                            (intoarce-legare-valoare 
                             (gaseste-legare (cdr pereche)
                              const-subst-list))
                            rezultat))))))


(defun generalizeaza-arbore (arb subst)
   "Generalizeaza un arbore de explicatie"
   ;; Intoarce arborele generalizat si lista substitutiilor
   ;; aplicate pentru aceasta
      (if (null arb)
      nil
      (let* ((const-subst-list (constr-lista-subst-const subst))
             (var-subst-list (constr-lista-subst-var subst
                               const-subst-list)))
         (list 
           (generalizeaza-arbore-aux arb () 'radacina 
             (append const-subst-list var-subst-list))
           var-subst-list))))

(defun generalizeaza-arbore-aux (arb father-id nod-id subst-list)
   ;; Functie auxiliara pentru generalizeaza-arbore. Traveseaza
   ;; arb in ordine si aplica subst tuturor nodurilor
   (setq arb (gen-nod arb nod-id subst-list))
   (if (not (frunza-p nod-id arb))
      (let ((rezultat arb))
         (dolist (fiu (nod-fii (obtine-nod nod-id arb)) rezultat)
            (setq rezultat
              (generalizeaza-arbore-aux rezultat nod-id fiu subst-list))))
      arb))



(defun gen-nod (arb nod-id subst-list)
   ;; Generalizeaza un nod. Intoarce arborele modificat
   (let ((info (nod-info (obtine-nod nod-id arb))))
      (act-info nod-id arb (gen-pred info subst-list))))

(defun gen-pred (sablon subst-list)
   "Generalizeaza un sablon "
   ;; Inlocuieste fiecare variabila sau constanta din
   ;; sablon cu legarea sa din subst-list, daca exista.
   ;; Sablon are forma 
   ;; (predicat argument [argument])
   ;; argument ==> (var nume) | constant
   ;; predicat ==> predicat-nume
   ;; Al doilea argument e optional
   
   (let* ((predicat (first sablon))
          (arg-prim (second sablon))
          (arg-secund (third sablon))
          (arg-prim-legare
           (intoarce-legare-valoare
            (gaseste-legare arg-prim subst-list)))
          (arg-secund-legare
           (intoarce-legare-valoare
            (gaseste-legare arg-secund subst-list)))
          (rezultat (list predicat 
                      (if (null arg-prim-legare)
                         arg-prim
                         arg-prim-legare))))
      (if (not (null arg-secund))
         (append rezultat
           (list (if (null arg-secund-legare)
                    arg-secund
                    arg-secund-legare)))
         rezultat)))      



(defun instanta (sablon1 sablon2)
   ;; testeaza daca sablon1 este o instanta a lui sablon2
  (not (equal (unifica sablon1 sablon2 ()) 'insucces)))
                                     
                                    
(defun filtreaza-arbore (arb parinte-id nod-id)
   ;; Elimina din arbore toate frunzele care sunt instante ale
   ;; nodurilor parinte. Aceasta se face pentru ca ele nu demonstreaza
   ;; parintele, ci doar il instantiaza.
   (cond ((and (frunza-p nod-id arb) 
               (instanta (nod-info (obtine-nod parinte-id arb))
                            (nod-info (obtine-nod nod-id arb))))
          (sterge-fiu parinte-id nod-id arb))
         (t (let ((rezultat arb))
               (dolist (fiu
                         (nod-fii (obtine-nod nod-id arb)) rezultat)
                  (setq rezultat (filtreaza-arbore rezultat nod-id fiu)))))))


         
;;////// Obtinerea unei reguli dintr-un arbore de justificare generalizat
         
(defun genereaza-regula (arb gen-subst)
   "Genereaza o regula dintr-o struct. de explicatie"
   ;; Noua regula va avea drept premize conjunctia frunzelor arborelui
   ;; si drept concluzie radacina acestuia
   (if (null arb)
      nil
      (let* ((lista-frunze (frunze arb))
             (lista-info-frunze 
              (mapcar #'(lambda (nod-id)
                          (nod-info (obtine-nod nod-id arb)))
                lista-frunze))
             (premize (if (> (length lista-info-frunze) 1)
                         (cons 'and lista-info-frunze)
                         lista-info-frunze))
             (concluzie (nod-info (obtine-nod 'radacina arb))))
         (make-regula :id (gensym)
           :premize premize
           :concluzie concluzie))))  



(defun op-concept (solutie)
   "Obtine o regula dintr-un concept prezentat ca exemplu"
   ;; Generalizeaza arborele de justificare si extrage din el
   ;; o noua regula. Intoarce regula obtinuta
   (if (null solutie)
      ()
      (let* ((arbore (solutie-arbore solutie))
             (subst (solutie-subst solutie))
             (arbore-filtrat (filtreaza-arbore arbore () 'radacina))
             (gen-arbore-res (generalizeaza-arbore arbore-filtrat subst))
             (gen-arbore (first gen-arbore-res))
             (gen-subst (second gen-arbore-res))
             (gen-rule (genereaza-regula gen-arbore gen-subst)))
         (format t "~&Regula obtinuta : ~a ~%" gen-rule)
         gen-rule)))


;; //////////////////////////////////////////////////////////////////////
;;
;; Exemplu de apel
;;
;; Functia ce operationalizeaza un concept (op-concept) primeste o solutie
;; (lista de substitutii si arbore de explicatie) si intoarce o regula
;; generata pe baza acestei solutii prin generalizarea arborelui.
;; Datorita presupunerii facute ca explicatia cea mai simpla este si cel
;; mai probabil adevarata, op-concept se va aplica solutie celei mai simple
;; (rezultat al functiei gaseste-expl). De exemplu, pentru baza de cunostinte
;; din BC.txt, un apel poate fi
;;
;; (op-concept (gaseste-expl '(cana obj1)))
;;
;; Reguyla obtinuta poate fi tiparita indentat, stocata in baza de cunostinte
;; cu ajutorul functiilor din bc-init.lsp, etc. Baza de cunostinte poate fi
;; salvata (cu noile reguli) pentru o folosire viitoare. 
;; 