;;/////////////////////////////////////////////////////////
;;
;; CAPITOLUL 5 & 8
;;
;; Substitutii
;; Fisier subst.lsp
;; 
;; ////////////////////////////////////////////////////////
             


;; /////////// Predicate
            
(defun varp (elem)
   ;; testeaza daca elem e variablila
   (and (listp elem)
        (equal (car elem) 'var)
        (equal (length elem) 2)))

(defun constant-p (elem)
   ;; testeaza daca ele este o constanta
   (atom elem))



;; //////// Manipularea legarilor

;; O lista de substitutii e o lista de asocieri de forma 
;; ( (var-name . legare) ... (var-name . legare))


(defun intoarce-legare (var lista-subst)
   ;; intoarce legarea lui var din lista-subst
   (assoc var lista-subst :test #'equal))

(defun intoarce-legare-valoare (legare)
   (cdr legare))

(defun gaseste-legare (var lista-subst)
   (let ((legare (intoarce-legare-valoare (intoarce-legare var lista-subst))))
      (if (varp legare)
         (gaseste-legare legare lista-subst)
         legare)))

(defun adauga-subst (var legare lista-subst)
   (acons var legare lista-subst))

(defun redenum-var-reg (asert)
   (if (REGULA-p asert)
      (make-regula :id (regula-id asert)
        :premize (redenum-variabile (regula-premize asert))
        :concluzie (redenum-rec (regula-concluzie asert)))
      asert))


(defun redenum-variabile (assert)
   ;; redenumeste var din asert cu nume unice, diferite
   (declare (special *lista-nume*))            
   (setq *lista-nume* nil)                 
   (redenum-rec assert))                 
     
(defun redenum-rec (exp)
   ;; redenumeste recursiv exp
   (declare (special *lista-nume*))       
   (cond ((constant-p exp) exp)
         ((varp exp) (redenum exp))
         (t (cons (redenum-rec (car exp)) 
              (redenum-rec (cdr exp))))))

(defun redenum (var)
   ;; redenumeste o variabila; cauta o redenumire anterioara
   ;; in lista-nume si daca nu gaseste genereaza un nume nou
   (declare (special *lista-nume*))
   (list 'var (or (cdr (assoc var *lista-nume* :test #'equal))
                  (let ((name (gensym)))
                     (setq *lista-nume* (acons var name *lista-nume*)) name))))

(defun aplic-subst (sablon lista-subst)
   ;; aplica substitutiile din lista-subst lui sablon
   (cond ((constant-p sablon) sablon)
         ((varp sablon)
          (let ((legare (gaseste-legare sablon lista-subst)))
             (cond (legare (aplic-subst (intoarce-legare-valoare legare) lista-subst))
                   (t sablon))))
         (t (cons (aplic-subst (car sablon) lista-subst)
                  (aplic-subst (cdr sablon) lista-subst))))) 



;; ////////// Diverse functii auxiliare

(defun invers-subst (lista-subst)
   "Inverseaza legarile din lista-subst"
   ;; Lista initiala
   ;;(((var v1) . (var w1)) .. ((var vn) . (var wn)))
   ;; lista finala  
   ;;(((var w1) . (var v1)) .. ((var wn) . (var vn)))
   (let ((resultat ()))
      (dolist (pereche lista-subst resultat)
         (setq resultat (acons (cdr pereche) (car pereche) resultat)))))


(defun sort-lista-subst (lista-subst)
   "Sorteaza lista-subst dupa numele variabilelor"
   (sort lista-subst #'string-lessp :key #'(lambda (x) (second (car x)))))



; /////////// Mapare substitutii

(defun map-subst (sablon-lista-subst lista-subst)
   "Compune doua substitutii"
   ;; sablon-lista-subst = lista-subst :
   ;; ( ((var v1) . obj1) .. ((var vn) . objn) )
   ;; lista-subst :
   ;; ( ((var v1) . (var w1)) .. ((var vn) . (var wn))) 
   ;;  Rezultat:
   ;; ((var w1) . obj1) ... ((var wn).  objn)))
   (let ((resultat ()))
      (dolist (pereche sablon-lista-subst resultat)
         (print pereche)
         (let* ((legare (intoarce-legare-valoare
                       (intoarce-legare (car pereche) lista-subst)))
                (legare-noua (cons legare (cdr pereche))))
            (if (and (not (null legare))
                     (not (member legare-noua resultat :test #'equal)))
               (setq resultat
                 (acons legare (cdr pereche) resultat)))))))
