
;;;////////////////////////////////////////////////////////
;;;
;;; CAPITOL 5 & 8
;;;  
;;; Initializare BC
;;; Fisier bc-init.lsp
;;;
;;;////////////////////////////////////////////////////////




;; ////////// Structuri de date si functii de manipulare ///

;;  TABELA hash 
(defvar *BC* ())
(setf *BC* (make-hash-table :size 20))

;;Toate asertiunile (reguli + fapte)
(defvar *LISTA-ASERTIUNI* ())

;; Toate regulile din BC
(defvar *LISTA-reguli* ())

;;Toate faptele din BC
(defvar *LISTA-FAPTE* ())
    
(defun adauga-regula-la-lista (id)
   ;;adauga o noua regula
   (setf *LISTA-reguli* (append *LISTA-reguli* (list id)))
   (setf *LISTA-ASERTIUNI* (append (list id) *LISTA-ASERTIUNI*)))

(defun adauga-fapt-la-lista (id)
   ;;adauga un nou fapt la lista fapte
   (setf *LISTA-FAPTE* (append *LISTA-FAPTE* (list id)))
   (setf *LISTA-ASERTIUNI* (append (list id) *LISTA-ASERTIUNI*)))

(defun adauga-asert-la-BC (asert)  
   ;; Adauga o asertiune la BC 
   (if (REGULA-p asert)
      (let ((id-reg (REGULA-id asert)))
         (if (not (member id-reg *LISTA-REGULI*))
            (progn
              (print asert)
              (setf (gethash id-reg *BC*) asert)
              (adauga-regula-la-lista id-reg))
            (format t "~a Deja in BC!~%" id-reg)))
      (let ((id-fapt (FAPT-id asert)))
         (if (not (member id-fapt *LISTA-FAPTE*))
            (progn
              (print asert)
              (setf (gethash id-fapt *BC*) asert)
              (adauga-fapt-la-lista id-fapt))
            (format t "~a Deja in BC!~%" id-fapt)))))
         
          
(defun obtine-asert (id)   
   (gethash id *BC*)) 

(defun obtine-asert-urm (regula-curenta)
   ;;obtine asertiunea (regula) urmatoare
   (second (member regula-curenta *LISTA-reguli*)))



;;;/////// I/O


(defun citeste-asert (input-stream)
   ;; citeste descrierea unei reguli de la input-stream
   (let ((linie (read input-stream)))
      (if (equal (first linie) 'REGULA)
         (make-REGULA :id (gensym)
           :premize (second linie)
           :concluzie (third linie))
         (make-FAPT :id (gensym)
          :info (second linie)))))
      
      
(defun scrie-regula (output-stream regula)
   ;; scrie regula la output stream
   (format output-stream "~&( REGULA ~a ~a~%~a~%)" 
     (REGULA-id regula) (REGULA-premize regula) (REGULA-concluzie regula) ))


(defun scrie-fapt (output-stream fapt)
   (format output-stream "~&( FAPT ~a ~a )" (FAPT-id fapt) (FAPT-info fapt)))
                                        

(defun scrie-asert (output-stream asert)
   (if (REGULA-p asert)
      (scrie-regula output-stream asert)
      (scrie-fapt output-stream asert)))




;;; /////// Initializare  BC 

(defun init-BC (nume-fisier-input)
   "Initializeaza BC din fisierul specificat"
   (clrhash *BC*)               ; pentru siguranta
   (setq *LISTA-ASERTIUNI* ())   
   (with-open-file (input-stream nume-fisier-input
                    :direction :input)
     (do ((i 0 (+ i 1)))
         ((eof-p input-stream)
          (format t "~&Incarcat ~d asertiuni~%" i))
        (adauga-asert-la-BC 
         (redenum-var-reg (citeste-asert
                           input-stream)))))
   (setf *LISTA-ASERTIUNI* (reverse (append (list 'sfarsit-BC) *LISTA-ASERTIUNI*))))


(defun tipareste-BC ()
   ;; tipareste BC
   (if (= (hash-table-count *BC*) 0)
      (format t "~&BC goala !~%")
      (mapcan #'(lambda (asert-id)
                  (if (not (equal asert-id 'sfarsit-BC)) 
                     (scrie-asert t (obtine-asert asert-id))))
        *LISTA-ASERTIUNI*))) 

(defun salveaza-BC-aux (nume-fisier-iesire)
   ;; salveaza BC in fisier
   (with-open-file (output-stream nume-fisier-iesire
                    :direction :output
                    :if-exists :supersede
                    :if-does-not-exist :create)
     (dolist (id *LISTA-ASERTIUNI*)
        (let ((asert (obtine-asert id))) 
           (if (not (equal id 'sfarsit-BC))
              (write-asert output-stream asert))))))
   
(defun salveaza-BC ()
   (format t "~&Salvare BC ~%")
   (salveaza-BC-aux "C:\\test\\BC.txt"))

(defun salveza-regula (regula)
   "Adauga o regula invatata la BC"
   (adauga-asert-la-BC regula))  


(defun init ()
   (setf *LISTA-ASERTIUNI* ())
   (setf *LISTA-REGULI* ())
   (setf *LISTA-FAPTE* ())
   (init-BC "c:\\CARTE\\PROGRAME\\CAPITOL5\\5.3\\bc.txt")
   (setf *LISTA-REGULI* (append *LISTA-REGULI* (list 'sfarsit-BC))))

(init)
   