
;;//////////////////////////////////////////////////////////////////
;;
;; CAPITOLUL 8
;;
;; Invatarea folosind spatiul versiunilor
;; Fisier vers.lsp
;;///////////////////////////////////////////////////////////////////



;; //////////////////// Reprezentarea exemplelor //////////////////

;; ///// Caracteristici
;; caracteristica = atribut + valoare

(defun make-caract (atribut valoare)
   (list atribut valoare))

(defun caract-atribut (caract)
   (first caract))

(defun caract-valoare (caract)
   (second caract))


;; ///// Dimensiuni
;; dimensiune = atribut + o multime de valori posibile

(defun make-DIMENS (atribut valori)
    (list atribut valori))

(defun DIMENS-atribut (dimens)
    (first dimens))

(defun DIMENS-valori (dimens)
   (second dimens))


;; /////// EXEMPLU 
;; EXEMPLU = id + caracteristici + clasa

(defun make-EXEMPLU (id caract clasa)
    (list id caract clasa))

(defun EXEMPLU-id (exemplu)
    (first exemplu))

(defun EXEMPLU-caract (exemplu)
    (second exemplu))

(defun EXEMPLU-clasa (exemplu)
    (third exemplu))


;; ///////// Initializarea dimensiunilor problemei

(defun dimensiuni ()
  (list (make-DIMENS 'ocupanti '(prof admin))
	(make-DIMENS 'etaj '(patru cinci))
	(make-DIMENS 'dept '(cs ee))))

(defun clase () '(da nu))



;; //////// Exemple de invatare

;; Un set de exemple pentru care algoritmul converge

(defun init-exemple-conv()
  (setq exemple
    (list (make-EXEMPLU '1 '((ocupanti prof) (etaj patru) (dept cs)) 'da)
          (make-EXEMPLU '2 '((ocupanti admin) (etaj cinci) (dept cs)) 'nu)
          (make-EXEMPLU '3 '((ocupanti prof) (etaj cinci) (dept cs)) 'da)
          (make-EXEMPLU '4 '((ocupanti prof) (etaj cinci) (dept ee)) 'nu))))

;; Un set de exemple pentru care algoritmul diverge

(defun init-exemple-diverg()
  (setq exemple
	(append exemple
		(list (make-EXEMPLU '5 '((ocupanti admin) (etaj cinci) (dept ee)) 'da)))))

  
;; ///////// Frontiere

(defun init-frontiere()
  (setq frontiere
	(list (list '())
	      (list (EXEMPLU-caract (first exemple))))))



;; ////////// Functii auxiliare /////////////////////////////////////
  
(defun gaseste-toti (lista test)
   "Gaseste toate elementele din lista care satisfac testul"
    (do ((elemente lista (rest elemente)) (rezultate nil))
	((null elemente) (reverse rezultate))
      (if (funcall test (first elemente))
	  (setq rezultate (adjoin (first elemente) rezultate)))))
 
(defun caract ()
   "Genereaza toate dimensiunile problemei"
   (mapcan #'(lambda (dim)
               (mapcar #'(lambda (valoare)
                           (list (DIMENS-atribut dim) valoare))
                 (DIMENS-valori dim)))
     (dimensiuni)))


(defun concept-eq (concept1 concept2)
   "Testeaza egalitatea a doua concepte"
   (and (every #'(lambda (caract)
                   (member caract concept2 :test #'equal))
          concept1)
        (every #'(lambda (caract)
                   (member caract concept1 :test #'equal))
          concept2)))

(defun singleton (frontiera)
   "Testeaza daca o frontiera are un singur element"
  (eq (length frontiera) 1))
  


;; ////////// Rafinarea frontierelor ////////////////////////////////

(defun rafineaza (exemplu general specific)
   "Rafineaza frontierele general si specific corespunzator exemplului"
  (if (eq (EXEMPLU-clasa exemplu) 'da)
      (simplifica-specific (filtreaza-general exemplu general)
				 (gen-specific exemplu specific))
      (simplifica-general (spec-general exemplu general)
				(filtreaza-specific exemplu specific))))
  


;; ///////// Generalizarea frontierei specifice (S)
  
(defun gen-specific (exemplu frontiera)
   "Generalizarea frontierei specifice"
   ;; se generalizeaza fiecare concept din ea pana cand este
   ;; consistent cu exemplul pozitiv 
   (mapcan #'(lambda (concept)
			(aux-gen-specific exemplu concept))
	    frontiera))
  
(defun aux-gen-specific (exemplu concept)
   (if (consistent (EXEMPLU-caract exemplu) concept)
	(list concept)
	(gen-specific exemplu (general concept))))
  

  
;; ////////  Specializarea frontierei generale (G)

(defun spec-general (exemplu frontiera)
   "Specializarea frontierei generale"
   ;; Se specializeaza fiecare concept din ea pana cand
   ;; nu mai este consistent cu exemplul negativ
  (mapcan #'(lambda (concept)
		(aux-spec-general exemplu concept))
	  frontiera))
  
(defun aux-spec-general (exemplu concept)
  (if (not (consistent (EXEMPLU-caract exemplu) concept))
      (list concept)
      (spec-general exemplu (special concept))))



;; /////////////// Filtrarea frontierelor
  
(defun filtreaza-general (exemplu general)
   "Sterge din frontiera generala toate conceptele care nu
sunt consistente cu exemplul pozitiv"
    (gaseste-toti general
	     #'(lambda (concept)
		(consistent (EXEMPLU-caract exemplu) concept))))
  

(defun filtreaza-specific (exemplu specific)
   "Sterge din frontiera specifica toate conceptele care sunt
consistente cu exemplul negativ"
    (gaseste-toti specific
	     #'(lambda (concept)
		(not (consistent (EXEMPLU-caract exemplu) concept)))))

  
;; /////////// Simplificarea frontierelor
  
(defun simplifica-specific (general specific)
   "Simplificarea frontierei specifice"
   ;; Se elimina toate conceptele care sunt mai generale decat 
   ;; alte concepte din acceasi frontiera si toate cele care nu sunt
   ;; mai specifice decat alte concepte din frontiera generala
  (setq specific (sterge-mai-general specific))
  (setq specific (sterge-nu-mai-specific general specific))
  (list general specific)) 

 
(defun simplifica-general (general specific)
   "Simplificarea frontierei generale"
   ;; Se elimina toate conceptele care sunt mai specifice decat
   ;; alte concepte din aceeasi frontiera si toate cele care nu sunt
   ;; mai generale decat alte concepte din frontiera specifica
  (setq general (sterge-mai-specific general))
  (setq general (sterge-nu-mai-general general specific))
  (list general specific))


;; Deletion of the concepts more general than others
(defun sterge-mai-general (lista)
   "Sterge din lista toate conceptele care sunt mai 
generale decat altele din aceeasi lista"
    (gaseste-toti lista
	     #'(lambda (concept)
		(not (mai-general-l concept lista)))))
  
(defun sterge-mai-specific (lista)
   "Sterge din lista toate conceptele care sunt mai 
specifice decat altele din aceeasi lista"
    (gaseste-toti lista
	     #'(lambda (concept)
		(not (mai-specific-l concept lista)))))
  

  
(defun sterge-nu-mai-specific (general specific)
   "Sterge din frontiera specifica toate conceptele care nu
sunt mai specifice decat altele din frontiera generala"
    (gaseste-toti specific
	     #'(lambda (concept)
		(mai-specific-l concept general))))
  
;; Delete concepts from the general frontiera that are not more general than
;; concepts from the specific frontiera
  
(defun sterge-nu-mai-general (general specific)
   "Sterge din frontiera generala toate conceptele care nu
sunt mai generale decat altele din frontiera specifica"
    (gaseste-toti general
	     #'(lambda (concept)
		(mai-general-l concept specific))))  



;; ////////// Implementarea relatiei MAI-GENERAL-DECAT

(defun mai-general (concept1 concept2)
   "Mai general decat"
    (every #'(lambda (caract)
			(not (null (member caract concept2 :test #'equal))))
	   concept1))


(defun mai-general-l (concept lista-concepte)
   "Intoarce t daca concept e mai general decat cel putin
unul din conceptele din lista"
    (do ((lista lista-concepte (rest lista)))
	((null lista) nil)
      (if (and (not (equal concept (first lista)))
	       (mai-general concept (first lista)))
	  (return t))))  


;; ////////// Implementarea relatiei MAI-SPECIFIC-DECAT

(defun mai-specific (concept1 concept2)
   "Mai specific decat"
  (mai-general concept2 concept1))
 
  
(defun mai-specific-l (concept lista-concepte)
   "Intoarce t daca concept e mai specific decat cel putin
unul din conceptele din lista"
    (do ((lista lista-concepte (rest lista)))
	((null lista) nil)
      (if (and (not (equal concept (first lista)))
	       (mai-specific concept (first lista)))
	  (return t))))
  

  
;; /////////// Consistenta
 
(defun e-in (atrib caract)
   "Cauta atrib in lista de caracteristici"
  (do ((lista caract (rest lista)))
      ((null lista) nil)
    (if (equal atrib (caar lista))  ;;; caar -> feature-atrib ???
	(return t))))
  
(defun consistent (caract-lista concept)
   "Testeaz dac conceptul este consistent cu exemplul 
reprezentat prin lista sa de caracteristici"
   (every #'(lambda (caract)
              (or 
                  (not (e-in (caract-atribut caract) caract-lista))
                  (member caract caract-lista :test #'equal)))
     concept))


  
;;////////////// Generalizare si specializare

;; Generalizarea unui concept se face prin stergerea unora din
;; caracteristicile sale
  
(defun general (concept)
   "Generalizeaza concept"
   ;; Intoarce toate generalizarile sale
    (mapcar #'(lambda (caract)
		(remove caract concept :test #'equal))
	    concept))
  
  
;; Specializarea unui concept se face prin adaugarea unor
;; caracteristici consistente cu conceptul
  
(defun special (concept)
   "Specializeaza concept"
   ;;Intoarce toate specializarile sale
    (mapcar #'(lambda (caract) (cons caract concept))
	    (gaseste-toti (caract)
		     #'(lambda (f)
			(and (consistent (list f) concept)
			     (not (member f concept :test #'equal)))))))
  
  


;; ////////////// Functia principala /////////////////////////////
  
(defun versiuni (exemplu frontiere)
   "Invata folosind spatiul versiunilor"
   (mapc #'(lambda (exemplu)
             (setq frontiere (rafineaza exemplu 
                              (first frontiere)
                              (second frontiere)))
             (format t "~%Ex: ~A~%G: ~A~%S: ~A~%"
               exemplu (first frontiere) (second frontiere))
             (if (and (singleton (first frontiere))
                      (singleton (second frontiere)))
                (if (concept-eq (first (first frontiere)) 
                     (first (second frontiere)))
                   (format t "Concept invatat: ~A~%" (first frontiere))))
             (if (and (null (first frontiere))
                      (null (second frontiere)))
                (print "Nu exista concept consistent cu exemplele prezentate!")))
     exemple))


;; /////////////// Exemple apel

;; Functia main grupeaza functiile prezentate anterior
;; si face prelucrarea celor doua seturi de exemple (convergent
;; si divergent)
;; Exemplu de apel 
;;
;; (main)
  
(defun main ()  
  (print "DIMENSIUNI")
  (DIMENSIUNI)
  (print "Initializare set exemple convergente")
  (init-exemple-conv)
  (print "Initializare frontiere")
  (init-frontiere)
  (print "Procesare exemple")
  (versiuni exemple frontiere)
  (print "Initializare set exemple divergente")
  (init-exemple-diverg) 
  (print "Initializare frontiere")
  (init-frontiere)
  (print "Procesare exemple")
  (versiuni exemple frontiere))