;;//////////////////////////////////////////////////////////////////
;;
;; 7 Planificare automata
;; 7.2 Un sistem de planificare liniara
;; 7.3 Planificare in lumea blocurilor
;;
;;/////////////////////////////////////////////////////////////////
;;
;; "planificare.lsp" - rezolvarea problemelor utilizand planificarea liniara 
;; (sistemul STRIPS)

(load "ansicl.lsp")

(defvar *OPERATORI* nil)

(defstruct operator
  (actiune nil)
  (preconditii nil)
  (lista-adaugari nil)
  (lista-eliminari nil))

(defun utilizeaza (lista-operatori)
  "lista-operatori devine multimea operatorilor problemei. Intoarce numarul operatorilor"
    (length (setf *OPERATORI* lista-operatori)))

(defparameter *op-scoala*
 `(,(make-operator 
       :actiune '(executa conduce-fiul-la-scoala)
       :preconditii '(fiu-acasa automobil-functioneaza)
       :lista-adaugari '(fiu-la-scoala)
       :lista-eliminari '(fiu-acasa))
   ,(make-operator 
       :actiune '(executa service-instaleaza-baterie)
       :preconditii 
        '(automobil-necesita-baterie 
         service-cunoaste-problema service-primeste-bani)
        :lista-adaugari '(automobil-functioneaza))
   ,(make-operator 
       :actiune '(executa spune-service-problema)
       :preconditii '(in-comunicare-cu-service)
       :lista-adaugari '(service-cunoaste-problema))
   ,(make-operator 
       :actiune 
           '(executa telefoneaza-lista-adaugari-service)
       :preconditii '(stie-numar-de-telefon)
       :lista-adaugari '(in-comunicare-cu-service))
   ,(make-operator :actiune '(executa cauta-numar)
       :preconditii '(are-carte-de-telefon)
       :lista-adaugari '(stie-numar-de-telefon))
   ,(make-operator :actiune '(executa plateste-service)
       :preconditii '(are-bani)
       :lista-adaugari '(service-primeste-bani)
       :lista-eliminari '(are-bani))))

(utilizeaza *op-scoala*)

;;=========================================================================

(defun strips (stare scopuri 
          &optional (*OPERATORI* *OPERATORI*))
  (sterge #'atom (satisfacere-scopuri 
                    (cons '(START) stare) scopuri nil)))

(defun satisfacere-scopuri (stare scopuri stiva-scopuri)
  (let ((stare-curenta stare))
    (when
       (and (every 
              #'(lambda (scop)
                  (setf stare-curenta 
                    (realizeaza-scop stare-curenta scop 
		 		 stiva-scopuri)))
              scopuri)
            (subsetp scopuri stare-curenta :test #'equal))
      stare-curenta)))

(defun realizeaza-scop (stare scop stiva-scopuri)
  (cond ((member scop stare :test #'equal) stare)
        ((member scop stiva-scopuri :test #'equal) nil)
        (t (some #'(lambda (operator)
                      (aplica-operator stare scop 
                         operator stiva-scopuri))
              (extrage scop *OPERATORI* #'potrivit-p)))))


(defun aplica-operator (stare scop operator stiva-scopuri)
  (let ((stare2 (satisfacere-scopuri stare
                    (operator-preconditii operator)
                    (cons scop stiva-scopuri))))
    (when stare2
      (append 
        (sterge 
           #'(lambda (x)
               (member x 
                  (operator-lista-eliminari operator) 
                  :test #'equal))
           stare2)
        (list (operator-actiune operator))
        (operator-lista-adaugari operator)))))

(defun potrivit-p (scop operator)
  (member scop (operator-lista-adaugari operator) 
     :test #'equal))

(defun extrage (element secventa functie-test)
  (delete nil 
    (mapcar 
       #'(lambda (i) 
            (when (funcall functie-test element) i))     
       secventa)))

(defun sterge (functie-test lista)
  (delete nil 
    (mapcar 
       #'(lambda (elem) 
            (unless (funcall functie-test elem) elem))    
       lista)))

(defun subsetp (subset set &rest args)
  (every 
    #'(lambda (element) 
         (apply #'member element set args)) subset))

(defun realizeaza-scop (stare scop stiva)
  (cond 
    ((memberg scop stare) stare)
    ((memberg scop stiva) nil)
    (t (some 
         #'(lambda (operator)
             (aplica-operator stare scop operator stiva))
         (operatori-potriviti scop stare)))))

(defun operatori-potriviti (scop stare)
"Intoarce o lista de operatori potriviti sortata dupa numarul de preconditii nesatisfacute in stare"
  (sort (extrage scop *OP* #'potrivit)
     #'(lambda (x y) 
          (< (criteriu x stare)(criteriu y stare)))))

(defun criteriu (op stare)
  (length (remove-if #'(lambda (x) (memberg x stare))
              (operator-preconditii op))))
(defun memberg (elem lista)
  (member elem lista :test #'equal))

(defun adaug-daca (test lista)
  (remove-if #'(lambda(x) (not (funcall test x))) lista))

(defun este-actiune (element)
  (and (consp element) 
       (or (equal (first element) 'executa)
           (equal (first element) 'START))))

;; Exemplu de apel
;; (strips '(fiu-acasa are-carte-de-telefon are-bani) '(fiu-la-scoala) *OPERATORI*) 

;;==========================================================================================
(defun strips (stareinit scopuri)
  (adaug-daca #'este-actiune 
     (caut-strips (list stareinit) nil 
        #'strategie-eur scopuri)))

;; Exemplu de apel
;; (strips '(fiu-acasa are-carte-de-telefon are-bani) '(fiu-la-scoala) )

(defun caut-strips (open closed strategie scopuri)
  (format t "~& CAUT: ~a" open)
  (cond ((null open) nil)
      ((subsetp scopuri (first open)) (first open))
      (t (caut-strips (funcall strategie 
                        (expand open closed)
                        (rest open) scopuri)
          (cons (first open) closed) strategie scopuri))))

(defun expand (open closed)
  (remove-if 
     #'(lambda (stare)
          (or (memberg stare open)(memberg stare closed)))
     (succ-strips (first open))))

(defun strategie-eur (stari-noi rest-stari scopuri)
  "Intoarce noua lista OPEN cu starile sortate in ordine crescatoare a numarului de conditii din stare care nu sunt in scopuri (un fel de distanta a starii fata de scopuri)"
  (sort (append stari-noi rest-stari)
     #'(lambda (x y) 
          (< (criteriu x scopuri)(criteriu y scopuri)))))

(defun criteriu (st scopuri)
  (length (remove-if #'(lambda (conditie)
                          (memberg conditie scopuri))
                     st)))

(defun succ-strips (stare)
  (mapcar 
     #'(lambda(op)
          (append 
             (remove-if 
                #'(lambda(x) (memberg x (operator-le op)))
                stare)
             (list (operator-actiune op))
             (operator-la op)))
     (operatori-aplicabili stare)))

(defun operatori-aplicabili (stare)
  "Intoarce o lista de operatori aplicabili in stare"
  (adaug-daca 
     #'(lambda (op) 
          (subset (operator-preconditii op) stare))
     *OP*))

;;=======================================================================

(defun constructor (blocuri)
  (let ((operatori))
    (dolist (a blocuri)
     (dolist (b blocuri)
       (unless (equal a b)
          (dolist (c blocuri)
              (unless (or (equal c a)(equal c b))
                 (push (mutare a b c) operatori)))
          (push (mutare a 'masa b) operatori)
          (push (mutare a b 'masa) operatori))))
    operatori))

(defun mutare (a b c)
  (make-operator 
     :actiune `(executa (muta ,a de pe ,b pe ,c))
     :preconditii `((liber ,a)(liber ,c)(,a pe ,b))
     :lista-adaugari (muta a b c)
     :lista-eliminari (muta a c b)))

(defun muta (a b c)
(if (eq b 'masa) `((,a pe ,c))  `((,a pe ,c)(liber ,b))))

(utilizeaza (constructor '(a b c)))


