;;//////////////////////////////////////////////////////////////////
;;
;; 4.4 Demonstrarea teoremelor prin respingere rezolutiva
;; 4.4.4 Rezolutie liniara cu utilizarea clauzelor ordonate
;;
;;/////////////////////////////////////////////////////////////////

;; "rezordonata.lsp" - implemetarea rezolutiei liniare cu utilizarea clauzelor 
;; ordonate 

(load "unificare")

(defun pozitiv-p (literal)
  "Intoarce T daca literal este un literal pozitiv."
  (and (consp literal) (eq (first literal) 'poz)))

(defun negat-p (literal)
  "Intoarce T daca literal este un literal negativ."
  (and (consp literal) (eq (first literal) 'neg)))

(defun incadrat-p (literal)
  "Intoarce T daca literal este un literal incadrat."
  (and (consp literal) (eq (first literal) 'inc)))

(defun literal (predicat &optional (semn 'poz))
  "Intoarce literalul (semn predicat)."
  (cons semn predicat))

(defun semn (literal)
  "Intoarce semnul unui literal."
  (first literal))

(defun semn-complementar (literal)
  "Intoarce semnul complementar sau nil daca literalul este incadrat."
  (cond ((eq (semn literal) 'poz) 'neg)
        ((eq (semn literal) 'neg) 'poz)))

(defun predicat (literal)
  "Intoarce predicatul unui literal."
  (rest literal))

;; remove-duplicates
(defun sterge-duplicate (clauza)
  "Intoarce clauza fara literali duplicati."
  (let (clauza1)
    (dolist (literal clauza (reverse clauza1))
      (unless (member literal clauza1 :test #'equal)
        (push literal clauza1)))))

(defun elimina (clauza)
  "Intoarce clauza din care au fost eliminati toti
  literalii incadrati neurmati de literali neincadrati."
  (do ((clauza1 (reverse clauza) (rest clauza1)))
      ((not (incadrat-p (first clauza1)))
       (reverse clauza1))))

;; delete :count 1
(defun sterge-literal (literal clauza)
  "Intoarce clauza din care a fost eliminat literalul."
  (do* ((C clauza (rest C))
        (L (first C) (first C))
        clauza-rezultat)
      ((or (equal L literal) (null C))
       (append clauza-rezultat (rest C)))
    (push L clauza-rezultat)))

(defun rezolvent-ordonat (C B)
  "Intoarce rezolventul ordonat al perechii de clauze C B."
  (let* ((CR (reverse C))
         (ultim-C (first CR))
         (literal-B (literal (predicat ultim-C)
                             (semn-complementar ultim-C))))
    (elimina
      (sterge-duplicate
        (append (reverse (cons (literal (predicat ultim-C) 'inc) (rest CR)))
                (sterge-literal literal-B B))))))

(defun rezolvent-redus (rezolvent)
  "Intoarce rezolventul redus al clauzei rezolvent
  sau nil in cazul in care acesta nu exista."
  (let ((ultim (first (last rezolvent)))
        literal)
    (elimina
      (when (and (negat-p ultim)
                 (member (setf literal
                               (literal (predicat ultim) 'inc)) rezolvent
                         :test #'equal))
        (reverse (rest (reverse (sterge-literal literal rezolvent))))))))

(defun rezolva-ordonat-p (C B)
  "Intoarce o substitutie daca clauzele ordonate C si B pot rezolva,
  T daca rezolva fara substitutie si nil in caz contrar."
  (do* ((ultim-C (first (last C)))
        (B-aux B (rest B-aux))
        (literal-B (first B-aux) (first B-aux))
        succes mgu)
       ((or (null B-aux) succes)
        (if succes (if mgu mgu t) nil))
    (when (eq (semn-complementar ultim-C) (semn literal-B))
      (multiple-value-setq
        (succes mgu)
        (unifica (predicat ultim-C) (predicat literal-B))))))

(defvar *AD-MAX* 0
  "Adincimea maxima admisa a clauzei vide.")

(defun pereche-clauze (C B adincime-C)
  "Intoarce o celula reprezentind o pereche de
  clauze impreuna cu adincimea clauzei centrale."
  (cons (cons C adincime-C) B))

(defun adincime-clauze (pereche)
  "Intoarce adincimea unei perechi de clauze."
  (cdar pereche))

(defun substituie-variabile-in-clauza (clauza mgu)
  "Intoarce claza in care variabilele au fost substituite."
  (mapcar #'(lambda (L)
              (literal (substituie-variabile (predicat L) mgu) (semn L)))
          clauza))

(defun respingere (C0 S &optional (*AD-MAX* 10))
  "Intoarce T daca clauza vida rezulta din clauza C0 si setul de clauze S,
  nil in caz contrar."
  (let ((FRONTIERA
          (delete nil
            (mapcar
              #'(lambda (B0i)
                  (let ((mgu (rezolva-ordonat-p C0 B0i)))
                    (when mgu
                      (if (eq mgu t)
                        (pereche-clauze C0 B0i 0)
                        (pereche-clauze
                          (substituie-variabile-in-clauza C0 mgu)
                          (substituie-variabile-in-clauza B0i mgu)
                          0)))))
              S))))
    (when FRONTIERA
      (do* ((FRONT FRONTIERA)
            (pereche (first FRONT) (first FRONT))
            Ri Ri-redus adincime SUCCES)
           ((or (null FRONT) SUCCES) SUCCES)
        (setf FRONT (rest FRONT))
        (when (< (setf adincime (adincime-clauze pereche)) *AD-MAX*)
          (setf Ri (rezolvent-ordonat (caar pereche) (rest pereche)))
          (setf Ri-redus (rezolvent-redus Ri))
          (when Ri-redus (setf Ri Ri-redus))
          (unless (when (null Ri) (setf SUCCES t))
            (mapc
              #'(lambda (Bik)
                  (let ((mgu (rezolva-ordonat-p Ri Bik)))
                    (when mgu
                      (if (eq mgu t)
                        (push (pereche-clauze Ri Bik (1+ adincime)) FRONT)
                        (push (pereche-clauze
                                (substituie-variabile-in-clauza Ri mgu)
                                (substituie-variabile-in-clauza Bik mgu)
                                (1+ adincime))
                              FRONT)))))
                  S)))))))


;;Exemplu de apel
(setq SSS 
  '( ((poz (p (*VAR* x))))
     ((neg (p (*VAR* x))) (poz (q (*VAR* x) a)))
     ((neg (q (*VAR* y) a))) ) )

(setq CC0 '((poz (p (*VAR* x)))))

(respingere CC0 SSS 10)


