;;////////////////////////////////////////////////////////////////
;;
;; 4 Logica si demonstrarea teoremelor
;; 4.3 Transformarea unei formule in forma clauzala
;; 4.3.2 Implementarea transformarii in forma clauzala
;;
;;////////////////////////////////////////////////////////////////

;; "transform.lsp" - transformarea unei formule bine formate intr-o lista de 
;; clauze

(defvar *lista-clauze* '())

(defun este-implicatie (formula)
" verifica daca formula este de forma: F -> G"
   (and (consp formula) 
        (= 3 (length formula)) 
        (equal (second formula) '->)))

(defun este-echivalenta (formula)
" verifica daca formula este de forma: F <-> G"
   (and (consp formula) 
        (= 3 (length formula)) 
        (equal (second formula) '<->)))

(defun substitutie (formula var var-nou)
" inlocuieste fiecare apritie a lui <var> cu <var-nou> in lista <formula>"
  (if (consp formula)
       (if (consp (car formula))
           (if (equal (car formula) var)
            (cons var-nou 
               (substitutie (cdr formula) var var-nou))
            (cons (substitutie (car formula) var var-nou)
               (substitutie (cdr formula) var var-nou)))
           (cons (car formula) 
               (substitutie (cdr formula) var var-nou)))
       formula))

(defun elimina-echivalente-si-implicatii (formula)
 (if (consp formula)
  (cond 
   ((este-implicatie formula)
    (list (list 'not 
             (elimina-echivalente-si-implicatii 
                                  (first formula)))
          'or (elimina-echivalente-si-implicatii 
                                  (third formula))))
   ((este-echivalenta formula)
    (list
      (list (list 'not (elimina-echivalente-si-implicatii    
                                  (first formula))) 
                  'or (elimina-echivalente-si-implicatii 
                                  (third formula)))
      'and
      (list (list 'not (elimina-echivalente-si-implicatii  
                                  (third formula)))
                  'or (elimina-echivalente-si-implicatii  
                                  (first formula)))))
   (t (cons (elimina-echivalente-si-implicatii 
                                  (first formula)) 
            (elimina-echivalente-si-implicatii 
                                  (rest formula))))) 
 formula))

(defun este-negatie (formula) 
" intoarce adevarat daca formula este de forma: NOT F, altfel intoarce fals"
   (and (consp formula) (= 2 (length formula)) 
        (equal (first formula) 'not)))

(defun este-Morgan (formula)
" verifica daca formula este de forma NOT ( F OR | AND G)"
   (and (este-negatie formula)
        (let ((form (second formula)))
           (and (= 3 (length form))
                (or (equal (second form) 'or)
                    (equal (second form) 'and))))))

(defun conector-negat (conector)
" transforma OR in AND si invers"
   (if (equal conector 'or)
      'and
      'or))

 (defun este-cuantif (formula)
" testeaza daca formula este de forma
( E | V  (*VAR <nume-var>))"
   (and (consp formula) (= 2 (length formula))
        (or (equal (first formula) 'E) 
            (equal (first formula) 'V))))

(defun cuantif-negat (cuantif)
" Transforma E n V si invers ntr-o expresie de forma:
(V | E (*VAR* x))"
   (if (equal (first cuantif) 'E)
      (cons 'V (rest cuantif))
      (cons 'E (rest cuantif))))

(defun este-lista-cuantif (formula)
" verifica daca argumentul constituie prefixul (lista de cunatificari de variabile) unei formule"
  (and (consp formula) (este-cuantif (first formula))))

(defun este-formula-cuantif (formula)
" testeaza daca formula este de forma:
((V | E (*VAR* <nume-var>))* F)"
   (and (consp formula) (= 2 (length formula))
        (este-lista-cuantif (first formula)))) 

(defun propaga-negatii (formula)
   (cond ((and (este-negatie formula) 
               (este-formula-cuantif (second formula)))
          (let ((lista-cuantif (caadr formula))
                (form (cadadr formula)))
             (list 
               (mapcar #'cuantif-negat lista-cuantif)
               (propaga-negatii (list 'not form))))) 
         ((este-Morgan formula) 
          (let ((form (second formula)))
             (list (propaga-negatii 
                        (list 'not (first form)))
                   (conector-negat (second form))
                   (propaga-negatii 
                        (list 'not (third form))))))
         ((and (este-negatie formula) 
               (este-negatie (second formula)))
          (propaga-negatii (second (second formula))))
         ((consp formula)
          (cons (propaga-negatii (first formula))
                (propaga-negatii (rest formula))))
         (t formula)))

(defun apare-in (var formula)
" ntoarce adevarat daca argumentul <var> apare cel putin o data in formula <formula>. n caz contrar rezultatul este nil"
   (cond ((equal var formula) T)
         ((consp formula) 
          (or (apare-in var (first formula))
              (apare-in var (rest formula))))
         (t nil)))

(defun uneste (formula)
" daca matricea unei formule este ea nsasi formata din prefix si matrice, uneste cele doua prefixe."
	(if (and (consp formula)
	         (este-formula-cuantif formula)
	         (este-formula-cuantif (second formula)))
	  (list (append (first formula) (caadr formula))  
            (cadadr formula))
	  formula))


(defun divide (formula)
" imparte o formula bine formata in prefix si matrice. Prefixul este elementul de pe prima pozie a formulei, daca acesta este o lista de variabile cunatificate, sau nil in caz contrar. Matricea este restul formulei"
   (if (este-formula-cuantif formula)
      (values (first formula) (second formula))
      (values '() formula)))

(defun regula (formula)
   (if (and (consp formula) (= (length formula) 3)
            (or (equal (second formula) 'and)
                (equal (second formula) 'or)))
      (let* ((termen1 (first formula))
             (termen2 (third formula))
             (conector (second formula))
             (rezult1 (uneste (regula-1-3-4 termen1)))
             (rezult2 (uneste (regula-1-3-4 termen2)))
             prefix1 prefix2 matrix1 matrix2 
             prefix matrix)
         (multiple-value-setq (prefix1 matrix1) 
               (divide rezult1))
         (multiple-value-setq (prefix2 matrix2) 
               (divide rezult2))
         (do ((p1 prefix1) (p2 prefix2) )
             ((or (null p1) (null p2))
              (setf prefix 
                 (append (append prefix p1) p2)))
            (let ((prim1 (car p1)) (prim2 (car p2)))
               (cond ((or
                        (and (equal prim1 prim2)
                             (equal (car prim1) 'V) 
                             (equal conector 'and))
                        (and (equal prim1 prim2)
                             (equal (car prim1) 'E) 
                             (equal conector 'or)))
                      (setf prefix 
                          (append prefix (list prim1)))
                      (setf p1 (rest p1))
                      (setf p2 (rest p2)))
                     ((equal (cadr prim1) (cadr prim2))
                        (let ((subst 
                                (list '*var* (gensym))))
                           (setf prefix 
                            (append prefix 
                              (list prim1 
                               (list (car prim2) subst))))
                           (setf matrix2 
                            (substitutie matrix2 
                               (cadr prim2) subst))
                           (setf p1 (rest p1))
                           (setf p2 (rest p2))))
                     (t
                      (let ((subst 
                             (list '*var* (gensym))))
                       (setf prefix 
                          (append prefix (list prim1)))
                       (setf p2 (substitutie p2 
                                    (cadr prim1) subst))
                       (setf matrix2 (substitutie matrix2  
                                      (cadr prim1) subst))
                       (setf p1 (rest p1)))))))
         (if (null prefix)
            (list matrix1 conector matrix2)
            (list prefix 
               (list matrix1 conector matrix2))))
      formula))   

(defun regula-1-3-4 (formula)
   (multiple-value-setq (prefix matrix) (divide formula))
   (if (not (null prefix))
      (uneste (list prefix (uneste (regula-1-3-4 matrix))))
      (regula matrix)))

(defun redenumeste-variabile-legate (formula aparitii)
 (cond ((este-formula-cuantif formula)
        (let ((lista-var (mapcar 'cadr (car formula)))
              (rezult formula) (arg aparitii))
          (dolist (item lista-var
                    (list (first rezult)
                      (redenumeste-variabile-legate 
                                    (cadr rezult) arg)))
            (if (member item arg :test #'equal)
               (let ((symb (list '*var* (gensym))))
                 (setf arg (cons symb (delete item arg)))
                 (setf rezult 
                    (substitutie rezult item symb)))
               (setf arg (cons item arg))))))
       ((consp formula)
        (cons (redenumeste-variabile-legate 
                         (first formula) aparitii)
              (redenumeste-variabile-legate 
                         (rest formula) aparitii)))
       (t formula)))
   
(defun make-function (symb lista-var)
" constrieste reprezentarea unui termen functional"
   (cons symb lista-var))                     


(defun elimin-E (formula)
   (if (este-formula-cuantif formula)
      (let* ((prefix (first formula))
            (matrice (second formula))
            (rezult-prefix '())
            (rezult-matrice matrice)
            (list-var '()))
         (do ((item prefix (rest item)))
             ((null item)
              (list rezult-prefix rezult-matrice))
            (if (equal (caar item) 'E)
              (setf rezult-matrice 
                    (substitutie rezult-matrice 
                          (cadar item)
                         (if (null list-var)
                            (gensym)
                            (cons (gensym) list-var))))
               (progn
                 (setf list-var 
                    (append list-var (cdar item)))
                 (setf rezult-prefix 
                    (append rezult-prefix 
                        (list (first item))))))))
      formula))

(defun este-distributie-st (formula)
" verifica daca formula este de forma: F OR (G AND H)"
   (and (consp formula) (= (length formula) 3)
        (equal (second formula) 'or) 
        (consp (third formula))
        (= (length (third formula)) 3) 
        (equal (second (third formula)) 'and)))

(defun este-distributie-dr (formula)
" verifica daca formula este de forma (F AND G) OR H"
   (and (consp formula) 
        (= (length formula) 3) 
        (equal (second formula) 'or)
        (consp (first formula)) 
        (= (length (first formula)) 3)
        (equal (second (first formula)) 'and)))                      

(defun distributie-repetata (formula)
   (let ((indicator nil) rezultat)
      (labels ((distributie (formula)
                (cond ((este-distributie-st formula)
                       (let* ((F (distributie 
                                     (first formula)))
                              (W (third formula))
                              (G (distributie (first W)))
                              (H (distributie (third W))))
                          (setf indicator t)
                          (list (list F 'or G) 'and 
                              (list F 'or H))))
                      ((este-distributie-dr formula)
                       (let* ((W (first formula))
                              (F (distributie (first W)))
                              (G (distributie (third W)))
                              (H (distributie 
                                     (third formula))))
                          (setf indicator t)
                          (list (list F 'or H) 'and 
                             (list G 'or H))))
                      ((consp formula) 
                       (cons (distributie (first formula))
                          (distributie (rest formula))))
                      (t formula))))
        (setf rezultat (distributie formula))
        (do ()
            ((null indicator) rezultat)
           (setf indicator nil)
           (setf rezultat (distributie rezultat))))))

(defun literal (formula semn)
" costruieste un literal din semn si predicat"
   (list semn formula))

(defun este-predicat-pozitiv (formula)
" verifica daca formula este un literal pozitiv"
     (and (consp formula) 
          (not (or (consp (first formula)) 
                   (equal (first formula) 'not)
                   (equal (first formula) 'or)
                   (equal (first formula) 'and)))))

(defun clauze (formula)
   (let ((clauza-curenta '()))
      (labels (( creeaza-lista-clauze (formula)
      (cond ((este-predicat-pozitiv formula) 
             (push (literal formula 'poz) clauza-curenta))
            ((este-negatie formula)
             (push (literal (second formula) 'neg) 
                clauza-curenta))
            ((equal formula 'and)
             (push (reverse clauza-curenta) 
                *lista-clauze*)
             (setf clauza-curenta '()))
            ((consp formula)
             (creeaza-lista-clauze (first formula))
             (creeaza-lista-clauze (rest formula)))   
            (t T))))
        (creeaza-lista-clauze formula)
        (push (reverse clauza-curenta) *lista-clauze*))))

(defun liniarizeaza (formula)
" are acelasi efect cu uneste, dar actioneaza recursiv pentru ntreaga formula"
   (cond ((este-formula-cuantif formula)
          (do* ((matrice-aux formula)
                (matrice formula (second matrice))
                (prefix '() (append prefix 
                                (first matrice-aux))))
               ((not (este-formula-cuantif matrice))
                (list prefix (liniarizeaza matrice)))
             (setf matrice-aux matrice)))
         ((consp formula)
          (cons (liniarizeaza (first formula))  
             (liniarizeaza (rest formula))))
         (t formula)))

 (defun transforma (formula)
   (setf *lista-clauze* '())
   (clauze 
     (distributie-repetata 
       (second 
         (elimin-E 
           (liniarizeaza 
             (regula-1-3-4
               (redenumeste-variabile-legate
                 (propaga-negatii
                   (elimina-echivalente-si-implicatii
                     formula)) 
                 '() )))))))
   (print (reverse *lista-clauze* )))

;;Exemple de apel
(transforma '( ((V (*var* x))(V (*var* y))(V (*var* z))) (((legat (*var* x) (*var* y) (*var* z)) 
                                   and
                                   (bic (*var z)) )
                                  ->
                                   (merg (*var* x) (*var* y))))  )

(transforma '(((V (*var x))) (((woman (*var* x)) and 
                      (not ( ((E (*var* y))) ((car (*var* y)) and
                                          (has (*var* x) (*var* y))))))
                     -> (loves luca (*var* x)))  )  )  

(transforma '(not ( ((E (*var* x))) 
                      ((car (*var* x)) and (has maria (*var* x))))) )         


(transforma '( ((V (*var* x)))
               ( ((E (*var* y))) 
                 ( (p (*var x) (*var* y)) and
                   ( (p (*var y*) (*var x)) or 
                     ( ((E (*var* z)))
                       (q (*var* y) (*var* z)))))))  ) 

    